# For bootstrap, we need unweighted data so that one unit in the table corresponds to one observation
# When necessary, figures are rounded
library(gnm)

boot.models <- function(data, indices, args) {
  library(gnm)
  medians <- args$medians
  fit.models <- args$fit.models
  cov.w <- function(y, w, y.mediansM=y, y.mediansF=y) {
    m <- medians[as.character(tab$M[tab$S==y]), "M", as.character(y.mediansM)]
    f <- medians[as.character(tab$F[tab$S==y]), "F", as.character(y.mediansF)]
    tot <- m + f
    w <- w[tab$S==y]
    tot.mean <- sum(tot * w)/sum(w)
    m.mean <- sum(m * w)/sum(w)
    f.mean <- sum(f * w)/sum(w)
    
    res <- cov.wt(cbind(Total=tot, M=m, F=f),
                  w, method="ML")
    cov <- c(res$cov)
    names(cov) <- outer(c("Total", "M", "F"), c("Total", "M", "F"), FUN=paste, sep=":")
    c(cov, "Total"=tot.mean, "M"=m.mean, "F"=f.mean)
  }
  
  tab <- args$tab
  tab$Freq <- round(tab$Freq)
  
  # Create a table from the indices - one index identifies an observation in the original table,
  # following the cumulative sum, from 1 to sum(tab)
  tab$Freq[tab$Freq > 0] <- tapply(tabulate(indices, nbins=sum(tab$Freq)),
                                   rep.int(1:nrow(tab), tab$Freq),
                                   sum)
  
  newmodels <- tryCatch(fit.models(tab), error=function(e) NULL)
  if(is.null(newmodels)) return(NA)
  
  years <- as.character(unique(tab$S))
  
  newcvs <- sapply(newmodels, simplify=FALSE, function(model) {
    cov <- sapply(years, cov.w, fitted(model))
    cv <- sqrt(cov["Total:Total",])/cov["Total",]
    cv
  })
  covsat <- sapply(years, cov.w, tab$Freq)
  cvsat <- sqrt(covsat["Total:Total",])/covsat["Total",]
  
  for(cv in newcvs)
    if(abs(cv[1]-newcvs[[1]][1]) > 1e-4) return(rep(NA, 12))
  
  cat(".")
  
  newcvs.lastyear <- c(cvsat[2], rev(sapply(newcvs, function(x) x[2])))
  newcv.firstyear <- newcvs[[1]][1]
  c(newcv.firstyear, cvsat[2] - newcv.firstyear,
    newcvs.lastyear, newcvs.lastyear[-1] - cvsat[2],
    diff(newcvs.lastyear))
}

fit.models <- function(tab) {
  # To check that constraints are correct, use coef(m)[!is.na(coef(m))]:
  # non-NA coefficients are those which are unconstrained
  m1num <- gnm(Freq ~ M*F + F0*(Mnum + I(Mnum^2)) + M0*(Fnum + I(Fnum^2)),
               data=subset(tab, Y == levels(Y)[1]), family=quasipoisson,
               constrain="^MQ.*:FInactive$|^MInactive:FQ", constrainTo=rep(0, 2*19))
  # Objective is to reproduce fitted table for first year rather than actually observed table
  # (this limits problems due to imprecise estimation of association)
  tab$Freq[tab$Y == levels(tab$Y)[1]] <- round(fitted(m1num))
  
  m1sat <- gnm(Freq ~ M*F, data=subset(tab, Y == levels(Y)[1]), family=quasipoisson)
  cons <- pickCoef(m1sat, "M|F", value=TRUE)
  
  m1indep <- gnm(Freq ~ M + F, data=subset(tab, Y == levels(Y)[1]), family=quasipoisson)
  pat <- sprintf("^(M|F)(%s)$", paste(levels(tab$M), collapse="|"))
  mnull <- gnm(Freq ~ M + F + Y, data=tab, family=quasipoisson, trace=FALSE,
               constrain=pat, constrainTo=pickCoef(m1indep, pat, value=TRUE))
  mstable <- gnm(Freq ~ M + F + Y + M:F, data=tab, family=quasipoisson, trace=FALSE,
                 constrain=names(cons), constrainTo=cons)
  mMF0 <- gnm(Freq ~ M + F + M0*Y + F0*Y + M:F, data=tab, family=quasipoisson, trace=FALSE,
              constrain=names(cons), constrainTo=cons)
  # Model equivalent to the above one in terms of inequalities, modulo quantile approximation
  # (but gives different fit statistics)
  mMF <- gnm(Freq ~ M*Y + F*Y + M:F, data=tab, family=quasipoisson, trace=FALSE,
             constrain=names(cons), constrainTo=cons)
  # With linear trend
  # mMFT <- gnm(Freq ~ M*Y + F*Y + M:F + T*(F0:M) + T*(M0:F), data=tab, family=quasipoisson, trace=FALSE,
  #             constrain=sprintf("^(%s|.*[^T]:(M|F)0TRUE)$", paste(names(cons), collapse="|")),
  #             constrainTo=c(cons, rep(0, 42)))
  mMFT <- gnm(Freq ~ M*Y + F*Y + M:F + Y*(F0:M) + Y*(M0:F), data=tab, family=quasipoisson, trace=FALSE,
              constrain=names(cons), constrainTo=cons)
  # For checking purposes: give the same models as above (as long as we constrain the
  # correct coefficients: be careful about their order!)
  # mMF2 <- gnm(Freq ~ M*Y + F*Y + M:F, data=tab, family=quasipoisson, trace=TRUE,
  #             constrain="^M.*:F.*", constrainTo=pickCoef(m1sat, "^M.*:F.*", value=TRUE))
  # mMFT2 <- gnm(Freq ~ M*Y + F*Y + M:F + T*(F0:M) + T*(M0:F), data=tab, family=quasipoisson, trace=TRUE,
  #              constrain="^M[^0].*:F[^0].*|.*[^T]:(M|F)0TRUE$",
  #              constrainTo=c(pickCoef(m1sat, "^M.*:F.*", value=TRUE), rep(0, 42)))
  # all.equal(fitted(mMF), fitted(mMF2), check.attributes = F)
  # all.equal(fitted(mMFT), fitted(mMFT2), check.attributes = F)
  
  # Check consistency of models:
  y1 <- tab$Y == levels(tab$Y)[1]
  # - they should all reproduce the row/column sums for first year
  for(m in list(mnull, mstable, mMF0, mMF, mMFT)) {
    stopifnot(all.equal(tapply(fitted(m)[y1], list(tab$M[y1]), sum),
                        tapply(tab$Freq[y1], list(tab$M[y1]), sum), tolerance=1e-4))
    stopifnot(all.equal(tapply(fitted(m)[y1], list(tab$F[y1]), sum),
                        tapply(tab$Freq[y1], list(tab$F[y1]), sum), tolerance=1e-4))
  }
  # - mnull and mstable should use first year row/column profiles for all years
  for(m in list(mnull, mstable)) {
    for(y in levels(tab$Y)) {
      stopifnot(all.equal(prop.table(tapply(fitted(m)[tab$Y==y], list(tab$M[tab$Y==y]), sum)),
                          prop.table(tapply(tab$Freq[y1], list(tab$M[tab$Y==y]), sum)), tolerance=1e-4))
      stopifnot(all.equal(prop.table(tapply(fitted(m)[tab$Y==y], list(tab$F[y1]), sum)),
                          prop.table(tapply(tab$Freq[y1], list(tab$F[y1]), sum)), tolerance=1e-4))
    }
  }
  # - mMF0, mMF and mMFT should reproduce observed number of inactive for all years
  for(m in list(mMF0, mMF, mMFT)) {
    stopifnot(all.equal(tapply(fitted(m), list(tab$M, tab$Y), sum)["Inactive",],
                        tapply(tab$Freq, list(tab$M, tab$Y), sum)["Inactive",], tolerance=1e-4))
    stopifnot(all.equal(tapply(fitted(m), list(tab$F, tab$Y), sum)["Inactive",],
                        tapply(tab$Freq, list(tab$F, tab$Y), sum)["Inactive",], tolerance=1e-4))
  }
  # - mMF and mMFT should reproduce observed row/column sums for all years
  for(m in list(mMF, mMFT)) {
    stopifnot(all.equal(tapply(fitted(m), list(tab$M, tab$Y), sum),
                        tapply(tab$Freq, list(tab$M, tab$Y), sum), tolerance=1e-4))
    stopifnot(all.equal(tapply(fitted(m), list(tab$F, tab$Y), sum),
                        tapply(tab$Freq, list(tab$F, tab$Y), sum), tolerance=1e-4))
  }
  # - mnull should have null association for all years
  for(y in levels(tab$Y)) {
    stopifnot(sum(abs(logmult:::lambda(tapply(fitted(mnull)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)))) < 1e-4)
  }
  # - mstable should reproduce first year proportions for all years
  for(y in levels(tab$Y)) {
    stopifnot(all.equal(prop.table(tapply(fitted(mstable)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)),
                        prop.table(tapply(tab$Freq[y1], list(tab$M[y1], tab$F[y1]), sum)), tolerance=1e-4))
  }
  # - mstable, mMF0 and mMFT should should reproduce first year associations for all years
  # (we cannot compare association direclty with observed data due to zeros, which are no reproduced exactly,
  #  so check against mstable whose proportions in first year have been checked above)
  for(y in levels(tab$Y)) {
    stopifnot(all.equal(logmult:::lambda(tapply(fitted(mstable)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)),
                        logmult:::lambda(tapply(fitted(mstable)[y1], list(tab$M[y1], tab$F[y1]), sum)), tolerance=1e-4))
    stopifnot(all.equal(logmult:::lambda(tapply(fitted(mMF0)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)),
                        logmult:::lambda(tapply(fitted(mstable)[y1], list(tab$M[y1], tab$F[y1]), sum)), tolerance=1e-4))
    stopifnot(all.equal(logmult:::lambda(tapply(fitted(mMF)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)),
                        logmult:::lambda(tapply(fitted(mstable)[y1], list(tab$M[y1], tab$F[y1]), sum)), tolerance=1e-4))
  }
  # - mstable, mMF0, mMF and mMFT should reproduce first year associations for all years among dual-earners
  # (we cannot compare association direclty with observed data due to zeros, which are no reproduced exactly,
  #  so check against mstable whose proportions in first year have been checked above)
  for(y in levels(tab$Y)) {
    stopifnot(all.equal(logmult:::lambda(tapply(fitted(mstable)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)[1:20, 1:20]),
                        logmult:::lambda(tapply(fitted(mstable)[y1], list(tab$M[y1], tab$F[y1]), sum)[1:20, 1:20]), tolerance=1e-4))
    stopifnot(all.equal(logmult:::lambda(tapply(fitted(mMF0)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)[1:20, 1:20]),
                        logmult:::lambda(tapply(fitted(mstable)[y1], list(tab$M[y1], tab$F[y1]), sum)[1:20, 1:20]), tolerance=1e-4))
    stopifnot(all.equal(logmult:::lambda(tapply(fitted(mMF)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)[1:20, 1:20]),
                        logmult:::lambda(tapply(fitted(mstable)[y1], list(tab$M[y1], tab$F[y1]), sum)[1:20, 1:20]), tolerance=1e-4))
    stopifnot(all.equal(logmult:::lambda(tapply(fitted(mMFT)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)[1:20, 1:20]),
                        logmult:::lambda(tapply(fitted(mstable)[y1], list(tab$M[y1], tab$F[y1]), sum)[1:20, 1:20]), tolerance=1e-4))
  }
  # - mMF0 and mMF should have equal associations for all years
  for(y in levels(tab$Y)) {
    stopifnot(all.equal(logmult:::lambda(tapply(fitted(mMF0)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)),
                        logmult:::lambda(tapply(fitted(mMF)[tab$Y==y], list(tab$M[tab$Y==y], tab$F[tab$Y==y]), sum)), tolerance=1e-4))
  }
  # - mstable, mMF0, mMF and mMFT should reproduce observed values for first year
  for(m in list(mstable, mMF0, mMF, mMFT)) {
    stopifnot(all.equal(fitted(m)[y1], tab$Freq[y1], tolerance=1e-4, check.attributes=FALSE))
  }
  # - mMFT should reproduce observed row and column profiles for the inactive for all years
  for(y in levels(tab$Y)) {
    stopifnot(all.equal(prop.table(tapply(fitted(mMFT)[tab$Y==y], list(tab$M[tab$Y==y]), sum))["Inactive"],
                        prop.table(tapply(tab$Freq[tab$Y==y], list(tab$M[tab$Y==y]), sum))["Inactive"],
                        tolerance=1e-4))
    stopifnot(all.equal(prop.table(tapply(fitted(mMFT)[tab$Y==y], list(tab$F[tab$Y==y]), sum))["Inactive"],
                        prop.table(tapply(tab$Freq[tab$Y==y], list(tab$F[tab$Y==y]), sum))["Inactive"],
                        tolerance=1e-4))
  }
  list(stable=mstable, MF=mMF, MFT=mMFT) # MF0=mMF0 isn't used
}


cvs <- list()
cl <- parallel::makePSOCKcluster(4)
for(c in countries) {
  cat(c, "\n")
  
  tab <- tabs[,, substr(dimnames(tabs)[[3]], 1, 2) == c]
  names(dimnames(tab)) <- c("M", "F", "S")
  
  tab <- as.data.frame(as.table(tab))
  tab$Freq <- round(tab$Freq) # Bootstrapping requires integer values
  
  tab$Y <- factor(substr(tab$S, 3, 4), levels=unique(substr(tab$S, 3, 4)))
  # Keep only first and last year for speed
  tab <- subset(tab, Y %in% c(levels(Y)[1], levels(Y)[length(levels(Y))]))
  tab$Y <- droplevels(tab$Y)
  # For linear trend
  # T <- ifelse(as.numeric(as.character(tab$Y)) < 70,
  #             2000 + as.numeric(as.character(tab$Y)),
  #             1900 + as.numeric(as.character(tab$Y)))
  # tab$T <- T - min(T)
  tab$F0 <- tab$F == "Inactive"
  tab$M0 <- tab$M == "Inactive"
  # For Schwartz (2010)
  #tab$ML <- ifelse(tab$M == "Inactive", 0, as.integer(tab$M))
  tab$Fnum <- as.numeric(ifelse(tab$F == "Inactive", 0, substr(tab$F, 2, 4)))
  tab$Mnum <- as.numeric(ifelse(tab$M == "Inactive", 0, substr(tab$M, 2, 4)))
  
  tab$liF <- medians[cbind(as.character(tab$M), "F", as.character(tab$S))]
  tab$liM <- medians[cbind(as.character(tab$F), "M", as.character(tab$S))]
  
  boots <- boot::boot(1:sum(tab$Freq, na.rm=TRUE), boot.models, R=500,
                      cl=cl, parallel="snow", ncpus=4,
                      args=list(tab=tab, fit.models=fit.models, medians=medians))
  cat("\n")
  cvs[[c]] <- boots
}
parallel::stopCluster(cl)


cis <- t(sapply(cvs, function(boot) {
  dat <- do.call(rbind, lapply(1:12, function(i) norm.ci(boot, index=i)[,-1, drop=F]))
  apply(dat, 1, function(x) paste0("(", paste(formatC(x, format="f", digits=3), collapse=", "), ")"))
}))
colnames(cis) <- c("CV First Year", "DCV", "S4", "S3", "S2", "S1", "DS3", "DS2", "DS1",
                   "DDS3", "DDS2", "DDS1")

save(cis, file="bootstrapped CVs.RData")